/* -----------------------------------------------------------------------------
*
*  npkey.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npkey.h"

#include "../npdata.h"
#include "npconsole.h"

#ifdef NP_MSW_
	#include <freeglut.h>
#endif
#ifdef NP_OSX_
	#include <GLUT/glut.h>
	#include <OpenGL/gl.h>
#endif
#ifdef NP_LINUX_
	#include <GL/freeglut.h>
#endif

#include "../npctrl.h"


void npKeyEvent (int key, int keyEventType);

//processes both the OS keyboard API and HID devices
//support for multiple devices mapped to multiple targets
//startup...
//console->mode = kNPconsoleCmd; //default keyboard context is rootHUD
//keyboard[i]->context = rootHUD //root HUD console
//
//active window is called (currently handled by GLUT)
//	allow keyboard or mouse to pick assigned windo
//		key/mouse (KM) pair assigned together with unique selection set
//			sub-selection / context within the window
//				ie. control two separate consoles
//
//event handler(s) called in order until event is handled
//  first handler could be ASCII Viewer user input while next is 3D Viewer, etc
//		allow for multiple simultaneous methods of user feedback, DB tree viewer
//
//	focus within the window is directed to the devices active context
//	rootHUD console and other node consoles... anywhere in tree
//
//event driven
//first sets modifiers states
//then converts API key code to ASCII and RAW
//next calls any event processing function(s) that are registered 
//key is converted to either ASCII

	
struct NPkeyMap
{
	int		id;
	int		type;
	char*	name;
	char*	note;
};
typedef struct NPkeyMap NPkeyMap;
typedef struct NPkeyMap *NPkeyMapPtr;


//------------------------------------------------------------------------------
void npInitKeyMap (void* dataRef)
{
	int i = 0;
	pData data = (pData) dataRef;

	static NPkeyMap base[] = {

	{ kNPGLboolean,			kNPGLboolean,	"GLboolean",		"openGL boolean" },
	{ kNPGLubyte,			kNPGLubyte,		"GLubyte",			"openGL ubyte" },
	{ kNPGLint,				kNPGLint,		"GLint",			"openGL int" },
	{ kNPGLfloat,			kNPGLfloat,		"GLfloat",			"openGL float" },

	{ kNPboolean,			kNPboolean,		"boolean",			"C boolean" },
	{ kNPubyte,				kNPubyte,		"ubyte",			"C ubyte" },
	{ kNPint,				kNPint,			"int",				"C int" },
	{ kNPfloat,				kNPfloat,		"float",			"openGL float" },

	// if both the constant and the data type are equal then it is a fundamental type
	// note can be a structure
	{ kNPfloatXYZA,			kNPfloatXYZA,	"NPfloatXYZA",		"openGL float x,y,z,angle" },
	{ kNPfloatXYZS,			kNPfloatXYZS,	"NPfloatXYZS",		"openGL x,y,z,scaler" },
	{ kNPfloatRGBA,			kNPfloatRGBA,	"NPfloatRGBA",		"openGL float r,g,b,a" },
	{ kNPfloatRGB,			kNPfloatRGB,	"NPfloatRGB",		"openGL float r,g,b" },
	{ kNPintXYZ,			kNPintXYZ,		"NPintXYZ",			"C int x,y,z" },
	{ kNPubyteRGBA,			kNPubyteRGBA,	"NPubyteRGBA",		"openGL ubyte r,g,b,a" },
	{ kNPboolXYZS,			kNPboolXYZS,	"NPboolXYZS",		"openGL boolean angle,x,y,z" },
	{ kNPboolXYZ,			kNPboolXYZ,		"NPboolXYZ",		"openGL boolean x,y,z" },
	{ kNPfloatXY,			kNPfloatXY,		"NPfloatXY",		"openGL float x,y" },
	{ kNPfloatXYZ,			kNPfloatXYZ,	"NPfloatXYZ",		"openGL float x,y,z" }
	};

}

//------------------------------------------------------------------------------
void npInitKey (void* dataRef)
{
	int i = 0;
	int j = 0;

	pData data = (pData) dataRef;
	pNPkey key = &data->io.key;

	key->modShift = 0;
	key->modCtrl = 0;
	key->modAlt = 0;
	key->modCommand = 0;

	for (i = 0; i < kNPkeyEventTypeSize; i++)
		for (j = 0; j < kNPkeyMapSize; j++)
			key->map[i][j] = 0;


	//the key map defines which command to execute for each key event
	//three types of fundamental keys, toggle, momentary, repeating
	//four event types generated, keyDown, KeyRepeat, KeyUp, KeyModifier

	//Fire keys, respond to keyDown only
	key->map[kKeyDown][kKeyCodeM]				= kNPcmdMenu;

	key->map[kKeyDown][kKeyCodeY]				= kNPcmdSegments;
	key->map[kKeyDown][kKeyCodeV]				= kNPcmdTexture;

	key->map[kKeyDown][kKeyCodeESC]				= kNPcmdFullscreen;
//	key->map[kKeyDown][kKeyCodeB]				= kNPcmdClearFrameBuffer;
	key->map[kKeyDown][kKeyCodeB]				= kNPcmdBackground;

	key->map[kKeyDown][kKeyCodeT]				= kNPcmdTool;
	key->map[kKeyDown][kKeyCodeX]				= kNPcmdSelectAxes;

	key->map[kKeyDown][kKeyCodeC]				= kNPcmdCamera;
	key->map[kKeyDown][kKeyCodeG]				= kNPcmdGrid;

	key->map[kKeyDown][kKeyCode4]				= kNPcmdSelectAll;
	key->map[kKeyDown][kKeyCodeTilda]			= kNPcmdSelectNone;
	key->map[kKeyDown][kKeyCodeSlash]			= kNPcmdSelectToggle; // Hijacked for database
	key->map[kKeyDown][kKeyCode1]				= kNPfileMapOne;
	key->map[kKeyDown][kKeyCode2]				= kNPfileMapTwo;
	key->map[kKeyDown][kKeyCode3]				= kNPfileMapThree;

	key->map[kKeyDown][kKeyCode5]				= kNPcmdPresetOne;
	key->map[kKeyDown][kKeyCode6]				= kNPcmdPresetTwo;
	key->map[kKeyDown][kKeyCode7]				= kNPcmdPresetThree;
//	key->map[kKeyDown][kKeyCode8]				= kNPcmdPresetFour;

//	key->map[kKeyDown][kKeyCodeC]				= kNPcmdCenter;
	key->map[kKeyDown][kKeyCodeBackSlash]		= kNPcmdScroll;
	key->map[kKeyDown][kKeyCodeF]				= kNPcmdFreeze;
	key->map[kKeyDown][kKeyCodeH]				= kNPcmdHide;

	key->map[kKeyDown][kKeyCodeO]				= kNPcmdGeometry;
	key->map[kKeyDown][kKeyCodeJ]				= kNPcmdTopo;

	key->map[kKeyDown][kKeyCodeI]				= kNPcmdTagMode;
	
	key->map[kKeyDown][kKeyCodeK]				= kNPfileSave;
	key->map[kKeyDown][kKeyCodeL]				= kNPfileOpen;
	key->map[kKeyDown][kKeyCodeP]				= kNPfileImport;	// kNPcmdPoints;

	key->map[kKeyDown][kKeyCodeSlash]			= kNPcmdViewer;

	key->map[kKeyDown][kKeyCodeU]				= kNPopenURL;


//	key->map[kKeyDown][kKeyCodeL]				= kNPcmdMesh;
	

	key->map[kKeyDown][kKeyCodeLeftBracket]		= kNPcmdSetpointLo;
	key->map[kKeyDown][kKeyCodeRightBracket]	= kNPcmdSetpointHi;

	key->map[kKeyDown][kKeyCodeSpace]			= kNPcmdPause;

	//Repeat keys, don't usually respond to keyUp
	key->map[kKeyDown][kKeyCodeN]				= kNPcmdNew;
	key->map[kKeyRepeat][kKeyCodeN]				= kNPcmdNew;
	
	key->map[kKeyDown][kKeyCodeDelete]			= kNPcmdDelete;
	key->map[kKeyRepeat][kKeyCodeDelete]		= kNPcmdDelete;

	key->map[kKeyDown][kKeyCodeBackSpace]		= kNPcmdDelete;
	key->map[kKeyRepeat][kKeyCodeBackSpace]		= kNPcmdDelete;

	key->map[kKeyDown][kKeyCodeTab]				= kNPcmdPin;
	key->map[kKeyRepeat][kKeyCodeTab]			= kNPcmdPin;

	key->map[kKeyDown][kKeyCodeReturn]			= kNPcmdNextBranch;
	key->map[kKeyRepeat][kKeyCodeReturn]		= kNPcmdNextBranch;
	key->map[kKeyDown][kKeyCodeNumPadEnter]		= kNPcmdNextBranch;
	key->map[kKeyRepeat][kKeyCodeNumPadEnter]	= kNPcmdNextBranch;

	key->map[kKeyDown][kKeyCodeComma]			= kNPcmdChannelDown;
	key->map[kKeyRepeat][kKeyCodeComma]			= kNPcmdChannelDown;
	
	key->map[kKeyDown][kKeyCodePeriod]			= kNPcmdChannelUp;
	key->map[kKeyRepeat][kKeyCodePeriod]		= kNPcmdChannelUp;

	key->map[kKeyDown][kKeyCodeMinus]			= kNPcmdColorDown;
	key->map[kKeyRepeat][kKeyCodeMinus]			= kNPcmdColorDown;
	
	key->map[kKeyDown][kKeyCodeEqual]			= kNPcmdColorUp;
	key->map[kKeyRepeat][kKeyCodeEqual]			= kNPcmdColorUp;

	key->map[kKeyDown][kKeyCodeZero]			= kNPcmdAlphaUp;
	key->map[kKeyRepeat][kKeyCodeZero]			= kNPcmdAlphaUp;

	key->map[kKeyDown][kKeyCode9]				= kNPcmdAlphaDown;
	key->map[kKeyRepeat][kKeyCode9]				= kNPcmdAlphaDown;

	key->map[kKeyDown][kKeyCode8]				= kNPcmdAlphaMode;
	key->map[kKeyRepeat][kKeyCode8]				= kNPcmdAlphaMode;

	key->map[kKeyDown][kKeyCodeR]				= kNPcmdRatio; //kNPcmdNormalMode;
	key->map[kKeyRepeat][kKeyCodeR]				= kNPcmdRatio; //kNPcmdNormalMode;

	

//	key->map[kKeyDown][kKeyCode]				= kNPcmdShaderMode;
//	key->map[kKeyRepeat][kKeyCode]				= kNPcmdShaderMode;


//	key->map[kKeyDown][kKeyCode]				= kNPcmdBiasUp;
//	key->map[kKeyRepeat][kKeyCode]				= kNPcmdBiasUp;
//	key->map[kKeyDown][kKeyCode]				= kNPcmdBiasDown;
//	key->map[kKeyRepeat][kKeyCode]				= kNPcmdBiasDown;

//	key->map[kKeyDown][kKeyCode]				= kNPcmdGainUp;
//	key->map[kKeyRepeat][kKeyCode]				= kNPcmdGainUp;
//	key->map[kKeyDown][kKeyCode]				= kNPcmdGainDown;
//	key->map[kKeyRepeat][kKeyCode]				= kNPcmdGainDown;

	//Momentary Keys are active while pressed and disable upon keyup
	//Arrow keys for X,Y,Z translate, rotate and scale of selected objects

	// translate
	key->map[kKeyDown][kKeyCodeA]				= kNPcmdXdecrease;
	key->map[kKeyUp][kKeyCodeA]					= kNPcmdXdecreaseOff;

	key->map[kKeyDown][kKeyCodeD]				= kNPcmdXincrease;
	key->map[kKeyUp][kKeyCodeD]					= kNPcmdXincreaseOff;

	key->map[kKeyDown][kKeyCodeW]				= kNPcmdYincrease;
	key->map[kKeyUp][kKeyCodeW]					= kNPcmdYincreaseOff;

	key->map[kKeyDown][kKeyCodeS]				= kNPcmdYdecrease;
	key->map[kKeyUp][kKeyCodeS]					= kNPcmdYdecreaseOff;

	key->map[kKeyDown][kKeyCodeE]				= kNPcmdZincrease;
	key->map[kKeyUp][kKeyCodeE]					= kNPcmdZincreaseOff;

	key->map[kKeyDown][kKeyCodeQ]				= kNPcmdZdecrease;
	key->map[kKeyUp][kKeyCodeQ]					= kNPcmdZdecreaseOff;

	// rotate
	key->map[kKeyDown][kKeyCodeLeft]			= kNPcmdRotateLeft;
	key->map[kKeyUp][kKeyCodeLeft]				= kNPcmdRotateLeftOff;

	key->map[kKeyDown][kKeyCodeRight]			= kNPcmdRotateRight;
	key->map[kKeyUp][kKeyCodeRight]				= kNPcmdRotateRightOff;

	key->map[kKeyDown][kKeyCodeUp]				= kNPcmdRotateUp;
	key->map[kKeyUp][kKeyCodeUp]				= kNPcmdRotateUpOff;

	key->map[kKeyDown][kKeyCodeDown]			= kNPcmdRotateDown;
	key->map[kKeyUp][kKeyCodeDown]				= kNPcmdRotateDownOff;

//	key->map[kKeyDown][kKeyCodeLeft]			= kNPcmdRotateCCW;
//	key->map[kKeyUp][kKeyCodeLeft]				= kNPcmdRotateCCWOff;

//	key->map[kKeyDown][kKeyCodeRight]			= kNPcmdRotateCW;
//	key->map[kKeyUp][kKeyCodeRight]				= kNPcmdRotateCWOff;

	// scale
	key->map[kKeyDown][kKeyCodeZ]				= kNPcmdZoomOn;
	key->map[kKeyUp][kKeyCodeZ]					= kNPcmdZoomOff;
}


//------------------------------------------------------------------------------
void npCloseKey (void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npKeyEvent (int key, int keyEventType)
{
	int command = 0;

	pData data = (pData) npGetDataRef();

	if (data->io.gl.hud.console.mode == kNPconsoleMenu)
		npConsoleKeyEvent (key, keyEventType, data);
	else
	{
		command = data->io.key.map[keyEventType][key];

		if (command != 0)
			npCtrlCommand (command, data);
	}
}
/*
enum
{
	kNPkeyNULL = 0,
	kNPkeyDown,
	kNPkeyUp,
	kNPkeyRepeat,
	kNPkeyDownSpecial,
	kNPkeyUpSpecial,
	kNPkeyRepeatSpecial
};
*/
enum
{
	kNPkeyCodeDefault = 0,
	kNPkeyCodeASCII,
	kNPkeyCodeGLUTspecial,
	kNPkeyCodeMSW,
	kNPkeyCodeOSX,
	kNPkeyCodeLinux,
	kNPkeyCodeHID
};

int npKeyRAW (int key, int keyCodeType);
//------------------------------------------------------------------------------
// Converts FLUT keycode to native keycode
int npKeyRAW (int key, int keyCodeType)
{
	int npKey = 0;
	
//	pData data = (pData) npGetDataRef();

//	npKeyRAW (key, kNPkeyDown, dataRef);
	if (keyCodeType == kNPkeyCodeASCII)
	{
		switch (key)
		{
			case 8  : npKey = kKeyCodeBackSpace; break;
			case 9  : npKey = kKeyCodeTab; break;
			case 13 : npKey = kKeyCodeReturn; break;
			case 27 : npKey = kKeyCodeESC; break;
			case 32 : npKey = kKeyCodeSpace; break;
			case 33 : npKey = kKeyCode1; break;
			case 34 : npKey = kKeyCodeRightQuote; break;  // double "
			case 35 : npKey = kKeyCode3; break;
			case 36 : npKey = kKeyCode4; break;
			case 37 : npKey = kKeyCode5; break;
			case 38 : npKey = kKeyCode7; break;
			case 39 : npKey = kKeyCodeRightQuote; break;  // single '
			case 40 : npKey = kKeyCode9; break;
			case 41 : npKey = kKeyCodeZero; break;
			case 42 : npKey = kKeyCode8; break;
			case 43 : npKey = kKeyCodeEqual; break;
			case 44 : npKey = kKeyCodeComma; break;
			case 45 : npKey = kKeyCodeMinus; break;
			case 46 : npKey = kKeyCodePeriod; break;	
			case 47 : npKey = kKeyCodeSlash; break;
			case 48 : npKey = kKeyCodeZero; break;
			case 49 : npKey = kKeyCode1; break;
			case 50 : npKey = kKeyCode2; break;
			case 51 : npKey = kKeyCode3; break;
			case 52 : npKey = kKeyCode4; break;
			case 53 : npKey = kKeyCode5; break;
			case 54 : npKey = kKeyCode6; break;
			case 55 : npKey = kKeyCode7; break;
			case 56 : npKey = kKeyCode8; break;
			case 57 : npKey = kKeyCode9; break;
			case 58 : npKey = kKeyCodeSemiColon; break;
			case 59 : npKey = kKeyCodeSemiColon; break;
			case 60 : npKey = kKeyCodeComma; break;
			case 61 : npKey = kKeyCodeEqual; break;
			case 62 : npKey = kKeyCodePeriod; break;	
			case 63 : npKey = kKeyCodeSlash; break;
			case 64 : npKey = kKeyCode2; break;
			case 65 : npKey = kKeyCodeA; break;
			case 66 : npKey = kKeyCodeB; break;
			case 67 : npKey = kKeyCodeC; break;
			case 68 : npKey = kKeyCodeD; break;
			case 69 : npKey = kKeyCodeE; break;
			case 70 : npKey = kKeyCodeF; break;
			case 71 : npKey = kKeyCodeG; break;
			case 72 : npKey = kKeyCodeH; break;
			case 73 : npKey = kKeyCodeI; break;
			case 74 : npKey = kKeyCodeJ; break;
			case 75 : npKey = kKeyCodeK; break;
			case 76 : npKey = kKeyCodeL; break;
			case 77 : npKey = kKeyCodeM; break;
			case 78 : npKey = kKeyCodeN; break;
			case 79 : npKey = kKeyCodeO; break;
			case 80 : npKey = kKeyCodeP; break;
			case 81 : npKey = kKeyCodeQ; break;
			case 82 : npKey = kKeyCodeR; break;
			case 83 : npKey = kKeyCodeS; break;
			case 84 : npKey = kKeyCodeT; break;
			case 85 : npKey = kKeyCodeU; break;
			case 86 : npKey = kKeyCodeV; break;
			case 87 : npKey = kKeyCodeW; break;
			case 88 : npKey = kKeyCodeX; break;
			case 89 : npKey = kKeyCodeY; break;
			case 90 : npKey = kKeyCodeZ; break;
			case 91 : npKey = kKeyCodeLeftBracket; break;
			case 92 : npKey = kKeyCodeBackSlash; break;
			case 93 : npKey = kKeyCodeRightBracket; break;
			case 94 : npKey = kKeyCode6; break;	
			case 95 : npKey = kKeyCodeMinus; break;
			case 96 : npKey = kKeyCodeTilda; break;
			case 97 : npKey = kKeyCodeA; break;
			case 98 : npKey = kKeyCodeB; break;
			case 99 : npKey = kKeyCodeC; break;
			case 100 : npKey = kKeyCodeD; break;
			case 101 : npKey = kKeyCodeE; break;
			case 102 : npKey = kKeyCodeF; break;
			case 103 : npKey = kKeyCodeG; break;
			case 104 : npKey = kKeyCodeH; break;
			case 105 : npKey = kKeyCodeI; break;
			case 106 : npKey = kKeyCodeJ; break;
			case 107 : npKey = kKeyCodeK; break;
			case 108 : npKey = kKeyCodeL; break;
			case 109 : npKey = kKeyCodeM; break;
			case 110 : npKey = kKeyCodeN; break;
			case 111 : npKey = kKeyCodeO; break;
			case 112 : npKey = kKeyCodeP; break;
			case 113 : npKey = kKeyCodeQ; break;
			case 114 : npKey = kKeyCodeR; break;
			case 115 : npKey = kKeyCodeS; break;
			case 116 : npKey = kKeyCodeT; break;
			case 117 : npKey = kKeyCodeU; break;
			case 118 : npKey = kKeyCodeV; break;
			case 119 : npKey = kKeyCodeW; break;
			case 120 : npKey = kKeyCodeX; break;
			case 121 : npKey = kKeyCodeY; break;
			case 122 : npKey = kKeyCodeZ; break;
			case 123 : npKey = kKeyCodeLeftBracket; break;
			case 124 : npKey = kKeyCodeBackSlash; break;
			case 125 : npKey = kKeyCodeRightBracket; break;
			case 126 : npKey = kKeyCodeTilda; break;
			case 127 : npKey = kKeyCodeDelete; break;
			default : break;
		}
	}

	// non-ASCII Glut constants
	if (keyCodeType == kNPkeyCodeGLUTspecial)
	{
		switch (key) 
		{
			case GLUT_KEY_F1 : npKey = kKeyCodeF1; break;
			case GLUT_KEY_F2 : npKey = kKeyCodeF2; break;
			case GLUT_KEY_F3 : npKey = kKeyCodeF3; break;
			case GLUT_KEY_F4 : npKey = kKeyCodeF4; break;
			case GLUT_KEY_F5 : npKey = kKeyCodeF5; break;
			case GLUT_KEY_F6 : npKey = kKeyCodeF6; break;
			case GLUT_KEY_F7 : npKey = kKeyCodeF7; break;
			case GLUT_KEY_F8 : npKey = kKeyCodeF8; break;
//			case GLUT_KEY_F9 : npKey = kKeyCodeF9; break;
//				case GLUT_KEY_F10 : npKey = kKeyCodeF10; break;
//				case GLUT_KEY_F11 : npKey = kKeyCodeF11; break;
//				case GLUT_KEY_F12 : npKey = kKeyCodeF12; break;
			case GLUT_KEY_LEFT : npKey = kKeyCodeLeft; break;
			case GLUT_KEY_RIGHT : npKey = kKeyCodeRight; break;
			case GLUT_KEY_UP : npKey = kKeyCodeUp; break;
			case GLUT_KEY_DOWN : npKey = kKeyCodeDown; break;
			case GLUT_KEY_PAGE_UP : npKey = kKeyCodePageUp; break;
			case GLUT_KEY_PAGE_DOWN : npKey = kKeyCodePageDown; break;
			case GLUT_KEY_HOME : npKey = kKeyCodeHome; break;
			case GLUT_KEY_END : npKey = kKeyCodeEnd; break;
//				case GLUT_KEY_INSERT : npKey = kKeyCode; break;	// not compatible with Mac
			default : break;
		}
	}

	//return converted keycode
	return npKey;
}

//------------------------------------------------------------------------------
// Process key events from GLUT
void npKeyGlut (int key, int x, int y, int keyTypeGlut, int modifiers)
{
	pData data = (pData) npGetDataRef();


	if (modifiers == GLUT_ACTIVE_SHIFT)		// move these GLUT commands to npGlut, debug zz
		data->io.key.modShift = true;
	else
		data->io.key.modShift = false;

	if (modifiers == GLUT_ACTIVE_CTRL)
		data->io.key.modCtrl = true;
	else
		data->io.key.modCtrl = false;

	if (modifiers == GLUT_ACTIVE_ALT)
		data->io.key.modAlt = true;
	else
		data->io.key.modAlt = false;

	//send raw ASCII text to console to process
	if (data->io.gl.hud.console.mode == kNPconsoleMenu)
	{
		switch (keyTypeGlut)
		{
			case kGlutKeyDown: npConsoleKeyEvent (key, kKeyDownASCII, data);
				break;
			case kGlutKeyDownSpecial:			// non-ASCII Glut constants
				switch (key) 
				{
					case GLUT_KEY_LEFT :	npConsoleKeyEvent (kKeyCodeLeft, kKeyDownSpecial, data); break;
					case GLUT_KEY_RIGHT :	npConsoleKeyEvent (kKeyCodeRight, kKeyDownSpecial, data); break;
					case GLUT_KEY_UP :		npConsoleKeyEvent (kKeyCodeUp, kKeyDownSpecial, data); break;
					case GLUT_KEY_DOWN :	npConsoleKeyEvent (kKeyCodeDown, kKeyDownSpecial, data); break;
				}
				break;
		}
		return;
	}

									//zz debug replace with keymap array, 
	//ASCII, GLUT, OSX, MSW, X11, HID
	//keyMode = kNPkeyRAW or kNPkeyASCII + special
	//keyEventType = kNPkeyDown //Up Repeat
	//keyType = QWERTY / ASCII / modifier / special
	//other HID devices, languages...? 
	//keyRepeatRate = 15
	//optionally load keymaps from CSV
	//default translate ASCII to local ASCII-friendly key code list

	switch (keyTypeGlut) 
	{
		case kGlutKeyDown:				// ASCII key codes converted to native
			npKeyEvent (npKeyRAW(key, kNPkeyCodeASCII), kKeyDown);
			break;
		case kGlutKeyUp:				// ASCII key codes converted
			npKeyEvent (npKeyRAW(key, kNPkeyCodeASCII), kKeyUp);
			break;
		case kGlutKeyDownSpecial:		// GLUT key codes for non-ASCII keys
			npKeyEvent (npKeyRAW(key, kNPkeyCodeGLUTspecial), kKeyDown);
			break;
		case kGlutKeyUpSpecial:			// non-ASCII Glut constants
			npKeyEvent (npKeyRAW(key, kNPkeyCodeGLUTspecial), kKeyUp);
			break;
		default : break;
	}
}

